package net.sf.corn.gate;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.LinkedList;
import java.util.List;

import net.sf.corn.exception.ExceptionBuilder;
import net.sf.corn.exception.ExceptionCommand;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * In order to execute a service operation that registered in the gate, the execution 
 * request simply need to be delegated to this class. To call a service operation, one can 
 * either use {@link #execute} methods of this class or ask to get proxy object of the service
 * by using {@link #getServiceProxy} methods. This class also offers {@link  #getActiveClientContext} method to access the active
 * client context, so service implementations can access the active context. 
 *  
 * @author Serhat Dirik
 *
 */
public class CallDelegator {

	private static class TLClientContextQueue extends
			ThreadLocal<LinkedList<IClientContext>> {

		LinkedList<IClientContext> context = null;

		@Override
		protected LinkedList<IClientContext> initialValue() {
			context = new LinkedList<IClientContext>();
			return context;
		}

		@Override
		public void set(LinkedList<IClientContext> value) {
			this.context = value;
		}

		@Override
		public LinkedList<IClientContext> get() {
			return context != null ? context : initialValue();
		}
	}

	private static Logger log = LoggerFactory
			.getLogger(CallDelegator.class);
	private static ExceptionCommand cmdLogError = new ExceptionCommand() {
		@Override
		public void execute(Exception e) {
			if (log.isDebugEnabled())
				log.error(e.getMessage(), e);
		}
	};

	private static TLClientContextQueue tlClientContextQueue = new TLClientContextQueue();

	private static Class<? extends IClientContext> localClientContext = SimpleLocalClientContext.class;

	/**
	 * Gives the registered local client context. This context is used when an internal 
	 * call initiated by an object that inside the same JVM as the service implementation.  
	 * {@link SimpleLocalClientContext} context implementation is used as the default local context implementation.
	 * @return local client context
	 */
	public static IClientContext getLocalClientContext() {
		IClientContext context = null;
		try {
			context = localClientContext.newInstance();
		} catch (InstantiationException | IllegalAccessException e) {
			e.printStackTrace();
		}
		return context;
	}
	/**
	 * This method sets local client context implementation default 
	 * @param context the context implementation to be used
	 * @throws IllegalAccessException  if the class or its nullary
     *               constructor is not accessible.
	 * @throws InstantiationException if context class represents an abstract class,
     *               an interface, an array class, a primitive type, or void;
     *               or if the class has no nullary constructor;
     *               or if the instantiation fails for some other reason.
	 */
	public static synchronized void setLocalClientContextClass(Class<? extends IClientContext> context) throws InstantiationException, IllegalAccessException {
		if (context == null) {
			throw new IllegalArgumentException("Null context class provided");
		}
		//Try to initialize
		context.newInstance();
		localClientContext = context;
	}

	/**
	 * Gives the active client context from the thread local queue. Service
	 * methods can access the actual context by using this method
	 * 
	 * @return caller context information
	 */
	public static IClientContext getActiveClientContext() {
		LinkedList<IClientContext> list = tlClientContextQueue.get();
		return list.peekLast();
	}

	private static void pushClientContext(IClientContext clientContext) {
		LinkedList<IClientContext> list = tlClientContextQueue.get();
		list.push(clientContext);
	}

	private static void removeClientContext() {
		LinkedList<IClientContext> list = tlClientContextQueue.get();
		list.pollLast();
		if (list.size() <= 0)
			tlClientContextQueue.set(null);
	}
	/**
	 * Analogous to use {@link #getServiceProxy(Class, IClientContext)} with null context parameter.
	 * @param clazz a java interface that annotated  with {@link GateService} 
	 * @return a service proxy object 
	 * @throws GateException
	 */
	public static <T> T getServiceProxy(Class<? extends T> clazz) throws GateException {
		return getServiceProxy(clazz, null);

	}

	/**
	 * Service proxy objects for services that defined as an interface which annotated with {@link GateService} 
	 * can be accessed by using this method.
	 * @param clazz a java interface that annotated  with {@link GateService} 
	 * @param context the context to be used for proxy object method calls. If null parameter sent, the delegator uses the active context.  
	 * @return a service proxy object 
	 * @throws GateException
	 */
	public static <T> T getServiceProxy(Class<? extends T> clazz, IClientContext context)
			throws GateException {
		if (clazz == null) {
			throw new IllegalArgumentException("class is null");
		}
		if(! clazz.isInterface()){
			throw new IllegalArgumentException("class is not an interface");
		}
		if (context == null) {
			context = CallDelegator.getActiveClientContext() == null ? getLocalClientContext()
					: CallDelegator.getActiveClientContext();
		}
		boolean inRegistry = false;
		ServiceDefinition sDef = new ServiceDefinition(clazz);
		if(Registry.getServiceDefinition(sDef.getName(),context) != null){
			inRegistry = true;
			sDef = Registry.getServiceDefinition(sDef.getName(),context);
		}

		if(!inRegistry){
			try {
				Registry.registerService(clazz);
			} catch (GateException e) {
				if (e.getCode() != GateException.CODE_REGISTRY_ALREADY_REGISTERED) {
					throw e;
				}
			}
		}
		@SuppressWarnings("unchecked")
		T proxy = (T) Proxy.newProxyInstance(clazz.getClassLoader(),
				new Class[] { IClientContextAccessor.class, clazz }, new ServiceProxy(sDef, context));
		return proxy;
	}
	
	/**
	 * Executes the given service operation with the given parameters. During the execution 
	 * the active client context is used. If there is no context is set, local context is used. 
	 * @param serviceName name of the service
	 * @param operationName name of the operation 
	 * @param parameters parameter list
	 * @return service response
	 */
	public static ServiceResponse execute(String serviceName,
			String operationName, Object... parameters) {
		return execute(new ServiceRequest(serviceName, operationName,
				parameters));
	}
	/**
	 * Executes the given service operation with the given parameters by using the given context.
	 * @param serviceName name of the service
	 * @param operationName name of the operation 
	 * @param parameters parameter list
	 * @return service response
	 */
	public static ServiceResponse execute(IClientContext context,
			String serviceName, String operationName, Object... parameters) {
		return execute(new ServiceRequest(serviceName, operationName,
				parameters, context));
	}

	/**
	 * Executes the given service request 
	 * @param request a service execution request
	 * @return service response
	 */
	public static ServiceResponse execute(ServiceRequest request) {
		ServiceResponse response = null;
		try {
			response = _execute(request);
		} catch (Exception e) {
			response = new ServiceResponse(ExceptionBuilder
					.with(GateException.class).cause(e).message(e.getMessage())
					.execute(cmdLogError).build()
					.code(GateException.CODE_SERVICE_UNEXPECTED));
		}
		return response;
	}

	private static ServiceResponse _execute(ServiceRequest request) {
		ServiceResponse response = null;
		// Get the classloader

		String returnType = null;
		ServiceDefinition serviceDef = null;
		OperationDefinition opDef = null;
		Method method = null;
		Object serviceInstance = null;
		Class<?> tReturnType = null;
		// Gather the context from the client
		IClientContext clientContext = request.getClientContext();
		// If the context is null, check if there is an active one exist
		// If the active context also null, use the default local one
		if (clientContext == null) {
			clientContext = CallDelegator.getActiveClientContext() == null ? getLocalClientContext()
					: CallDelegator.getActiveClientContext();
			//Push headers
			for(String key:request.headerKeySet()){
				clientContext.addToRequestHeader(key, request.getFromHeader(key));
			}
		}
		try {
			clientContext = wrapIfNecessary(clientContext);
		} catch (GateException e) {
			log.error("An error occured while wrapping the web context!", e);
		}

		try {
			ExecutionTarget execTarget = Registry.getExecutionTarget(
					clientContext, request.getServiceName(),
					request.getOperationName(), request.getParameters());
			// Get the service definition
			serviceDef = execTarget.getServiceDef();
			opDef = execTarget.getOperationDef();
			method = execTarget.getMethod();
			serviceInstance = execTarget.getServiceInstance();
			// Find if transformation is required for the return type
			returnType = method.getReturnType().getCanonicalName();
			tReturnType = findType(returnType);
			if(tReturnType == null){
				tReturnType = method.getReturnType();
			}

		} catch (GateException e) {
			response = new ServiceResponse(e);
			return response;
		} catch (ClassNotFoundException e) {
			response = new ServiceResponse(ExceptionBuilder
					.with(GateException.class).cause(e)
					.fill("Return type %1! is an unknown type '", returnType)
					.execute(cmdLogError).build()
					.code(GateException.CODE_GENERIC_CLASS_NOT_FOUND));
			return response;
		} catch (Exception e) {
			Exception retException = ExceptionBuilder
					.with(GateException.class, e)
					.fill("An exception occured while preparing execution target of the service %1!.%2!",
							request.getServiceName(),
							request.getOperationName()).execute(cmdLogError)
					.build().code(GateException.CODE_SERVICE_UNEXPECTED);
			response = new ServiceResponse(retException);
			return response;
		}

		// Set thread context values
		pushClientContext(clientContext);
		// Find filters
		List<ServiceFilter> filters = Registry.findFilters(
				serviceDef.getName(), request.getOperationName());
		// Notifications and advice executions
		for (ServiceFilter filter : filters) {
			 
			List<IExecutionAdvice> advices = filter.getAdvice();
			for (IExecutionAdvice advice : advices) {
				try {
					advice.executeBeforeMethod(request, serviceDef, opDef,
							serviceInstance, method);
				} catch (Exception ex) {
					response = new ServiceResponse(ex);
					removeClientContext();
					return response;
				}
			}
		}

		// Execute the method
		Object result = null;
		try {
			result = method.invoke(serviceInstance, request.getParameters());
			// Prepare response
			if(result != null){
				response = new ServiceResponse(result);
			}else{
				response = new ServiceResponse(result,tReturnType);
			}
		} catch (IllegalAccessException | IllegalArgumentException e) {
			Exception respEx = ExceptionBuilder
					.with(GateException.class, e)
					.execute(cmdLogError)
					.fill("An exception occured while executing the service %1!.%2!",
							request.getServiceName(),
							request.getOperationName()).build()
					.code(GateException.CODE_SERVICE_UNEXPECTED);
			response = new ServiceResponse(respEx);

		} catch (Exception e) {
			response = new ServiceResponse(e.getCause(), isExpectedException(
					method, e.getCause()));
		}

		//Push Response headers
		for(String key:clientContext.responseHeaderKeySet()){
			response.addToHeader(key, clientContext.getFromResponseHeader(key));
		}
		
		// Call after methods
		// Notifications and advice executions
		for (ServiceFilter filter : filters) {
			List<IExecutionAdvice> advices = filter.getAdvice();
			for (IExecutionAdvice advice : advices) {
				try {
					response = advice.executeAfterMethod(request, response, serviceDef, opDef, serviceInstance, method);
				} catch (Exception ex) {
					response = new ServiceResponse(ex);
					removeClientContext();
					return response;
				}
			}
		}

		// Reset context values
		removeClientContext();

		return response;
	}

	
	private static Class<?> findType(String className)
			throws ClassNotFoundException {
		switch (className.trim().toLowerCase()) {
		case "void":
			return Void.class;
		case "int":
			return int.class;
		case "byte":
			return byte.class;
		case "short":
			return short.class;
		case "long":
			return long.class;
		case "float":
			return float.class;
		case "double":
			return double.class;
		case "boolean":
			return boolean.class;
		case "char":
			return char.class;
		default:
			return null;
		}
	}

	/**
	 * Returns true if the {@link java.lang.reflect.Method Method} definition on
	 * the service is specified to throw the exception contained in the
	 * InvocationTargetException or false otherwise. NOTE we do not check that
	 * the type is serializable here. We assume that it must be otherwise the
	 * application would never have been allowed to run.
	 * 
	 * @param serviceIntfMethod
	 *            the method from the RPC request
	 * @param cause
	 *            the exception that the method threw
	 * @return true if the exception's type is in the method's signature
	 */
	private static boolean isExpectedException(Method serviceIntfMethod,
			Throwable cause) {
		assert (serviceIntfMethod != null);
		assert (cause != null);

		Class<?>[] exceptionsThrown = serviceIntfMethod.getExceptionTypes();
		if (exceptionsThrown.length <= 0) {
			// The method is not specified to throw any exceptions
			//
			return false;
		}

		Class<? extends Throwable> causeType = cause.getClass();

		for (Class<?> exceptionThrown : exceptionsThrown) {
			assert (exceptionThrown != null);

			if (exceptionThrown.isAssignableFrom(causeType)) {
				return true;
			}
		}

		return false;
	}

	private static IClientContext wrapIfNecessary(IClientContext context)
			throws GateException {
		if (context == null) {
			throw new IllegalArgumentException("Context is null!");
		}
		Class<? extends WrapperClientContext> contextWrapper = Registry
				.getContextWrapper();
		if (contextWrapper != null) {
			try {
				Constructor<? extends WrapperClientContext> constructor = (Constructor<? extends WrapperClientContext>) contextWrapper
						.getConstructor(IClientContext.class);
				return constructor.newInstance(context);
			} catch (InstantiationException | IllegalAccessException
					| IllegalArgumentException | InvocationTargetException
					| NoSuchMethodException | SecurityException e) {
				throw ExceptionBuilder
						.with(GateException.class, e)
						.message(e.getMessage())
						.execute(cmdLogError)
						.build()
						.code(GateException.CODE_GENERIC_CONTEXT_INITIALIZATION);
			}
		} else {
			return context;
		}
	}
}
