package net.sf.corn.gate.jsonrpc;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import net.sf.corn.converter.ConversionException;
import net.sf.corn.converter.ParsingException;
import net.sf.corn.converter.json.JsTypeComplex;
import net.sf.corn.converter.json.JsTypeObject;
import net.sf.corn.converter.json.JsTypeSimple;
import net.sf.corn.converter.json.JsonConverter;
import net.sf.corn.converter.json.JsonStringParser;

import org.springframework.beans.ConversionNotSupportedException;


/**
 * &nbsp;&nbsp;This class represents a response from a remote json-rpc service. The
 * response of a service call may contain a result or an error. In any cases
 * call responses contains a message identifier which is the same as incoming
 * message for message correlation on the caller side. The response may contain
 * a header message part which contains additional fields beside error or
 * result. The following Javascript class represents a response wrapper which
 * contains an operation result within it.<br>
 * <code>
 *  {<br>
 *   &nbsp;"jsonrpc":"2.0",<br>
 *   &nbsp;"id":"64f4f4b0-79ee-45f0-9b47-77836bc74ce8",<br>
 *   &nbsp;"header":{ <br>
 *     &nbsp;&nbsp;"headerField1": "headerVal",<br>
 *     &nbsp;&nbsp;"headerField2": "headerVal",<br>
 *   &nbsp;},<br>
 *   &nbsp;"result":"resultObject"<br>
 *  }<br>
 * </code> 
 * 
 * <br>
 * 
 * The following Javascript class represents a response which contains an error within it.<br>
 * <code>
 *   {<br>
 *   &nbsp;"id":"C3F236B3-D450-0001-B8E3-4F8D528C5390",<br>
 *   &nbsp;"error":<br>
 *   	&nbsp;&nbsp;{"name":"org.sr.server.AuhenticationException",<br>
 *   	 &nbsp;&nbsp;"code":491,<br>
 *       &nbsp;&nbsp;"message":"Authentication is required for  securityTestBean.authenticationRequiredMethod calls"<br>
 *       &nbsp;&nbsp;}<br>
 *    }
 * </code>
 * <hr>
 * <h2>Message Parts</h2>
 * <h3>id</h3>
 * &nbsp;This part should contain a unique message identifier for the message
 * correlation. Results of operations will be sent to the clients by using same
 * identifier as requests. This would help clients to correlate their responses
 * from server.
 * <h3>header</h3>
 * &nbsp; The header part of the message contains additional fields that sent by
 * the server beside parameters.
 * <h3>result</h3>
 * &nbsp;The result of the operation. In the case of an exception raises, the
 * wrapper class wouldn't contains result.
 * <h3>error</h3>
 * &nbsp;The wrapper class contains error part only if an exception raises from
 * the service. Error object has three attributes as errorCode,errorMsg and
 * errorName.
 * 
 * @author Serhat Dirik
 * 
 */
public class JsonRpcResponse implements Serializable {

	private static final long serialVersionUID = -6349761749792602787L;
	public static final String RESPONSEID_ATTRIBUTE = "id";
	public static final String HEADER_ATTRIBUTE = "header";
	public static final String RESULT_ATTRIBUTE = "result";
	public static final String ERROR_ATTRIBUTE = "error";
	public static final String ERRORNAME_ATTRIBUTE = "name";
	public static final String ERRORMESSAGE_ATTRIBUTE = "message";
	public static final String ERRORCODE_ATTRIBUTE = "code";
	public static final String ERRORDATA_ATTRIBUTE = "data";
	public static final String JSONRPC_ATTRIBUTE = "jsonrpc";
	public static final int    ERR_CODE_PARSE_ERR_NOT_WELL_FORMED = -32700; //Parse error	Invalid JSON was received by the server.An error occurred on the server while parsing the JSON text.
	public static final int    ERR_CODE_PARSE_ERR_UNSUPPORTED_ENCODING = -32701; //unsupported encoding
	public static final int    ERR_CODE_PARSE_ERR_INVALID_CHARACTER = -32702;//invalid character for encoding
	public static final int    ERR_CODE_SERVER_ERR_INVALID_REQUEST_OBJECT = -32600; //Invalid Request	The JSON sent is not a valid Request object. 
	public static final int    ERR_CODE_SERVER_ERR_METHOD_NOT_FOUND = -32601; //Invalid Request	The JSON sent is not a valid Request object.
	public static final int    ERR_CODE_SERVER_ERR_INVALID_PARAMETERS = -32602; //Invalid params	Invalid method parameter(s).
	public static final int    ERR_CODE_SERVER_ERR_INTERNAL_ERROR = -32603; //Internal error	Internal JSON-RPC error.
	public static final int    ERR_CODE_APPLICATION_ERROR = -32500; //Application Error
	public static final int    ERR_CODE_SYSTEM_ERROR = -32400; //system Error
	public static final int    ERR_CODE_TRANSPORT_ERROR = -32300; //transport Error
	
	private JsTypeComplex enclosedResponse = new JsTypeComplex();
	private transient JsonConverter converter = new JsonConverter();
	
	@SuppressWarnings("unused")
	private JsonRpcResponse() {
	}
	
	/**
	 * Constructs a new instance from a json string 
	 * @param jsonString a json syntaxed string that contains required fields
	 * @throws ConversionException
	 * @throws ParsingException
	 */
	JsonRpcResponse(String jsonString) throws ConversionException, ParsingException{
		JsTypeComplex json_resp = (JsTypeComplex) JsonStringParser.parseJsonString(jsonString);
		setResponse(json_resp);
		if(!enclosedResponse.has(JSONRPC_ATTRIBUTE)){
			setJsonrpcVersion("2.0");
		}
	}
	
	/**
	 * Create a JSONResponse wrapper java class to the given complex json object
	 * 
	 * @param response complex typed json object
	 * @throws ConversionException
	 */
	JsonRpcResponse(JsTypeComplex response) throws ConversionException {
		setResponse(response);

	}
	
	/**
	 * Creates a response wrapper class for the given result for the given
	 * message id
	 * 
	 * @param id
	 * @param result
	 * @throws ConversionNotSupportedException 
	 * @throws ConversionException 
	 * @throws JSONException
	 */
	JsonRpcResponse(String id, JsTypeObject result) 
			throws ConversionException {
		if(id != null) enclosedResponse.put(RESPONSEID_ATTRIBUTE, new JsTypeSimple(JsTypeObject.quote(id),null));
		enclosedResponse.put(RESULT_ATTRIBUTE, result);
			setJsonrpcVersion("2.0");
	}

	/**
	 * Creates a response wrapper class which contains the given error
	 * 
	 * @param id
	 * @param errorName
	 * @param errorMessage
	 * @param errorCode
	 * @throws ConversionNotSupportedException 
	 * @throws ConversionException 
	 * @throws JSONException
	 */
	JsonRpcResponse(String id,String errorName,
			String errorMessage,int errorCode) {
		this(id,errorName,errorMessage,errorCode,null);
	}
	/**
	 * Creates a response wrapper class which contains the given error
	 * 
	 * @param id
	 * @param errorName
	 * @param errorMessage
	 * @param errorCode
	 * @param errorData
	 * @throws ConversionNotSupportedException 
	 * @throws ConversionException 
	 * @throws JSONException
	 */
	JsonRpcResponse(String id,String errorName,
			String errorMessage,int errorCode,Object errorData) {
		try {
			if(id != null) enclosedResponse.put(RESPONSEID_ATTRIBUTE, new JsTypeSimple(JsTypeObject.quote(id),null));
			JsTypeComplex error = new JsTypeComplex();
			error.putObject(ERRORMESSAGE_ATTRIBUTE,  new JsTypeSimple(JsTypeObject.quote(errorMessage),null));
			error.putObject(ERRORNAME_ATTRIBUTE,  new JsTypeSimple(JsTypeObject.quote(errorName),null));
			error.putObject(ERRORCODE_ATTRIBUTE,  errorCode);
			if(errorData != null){
				error.putObject(ERRORDATA_ATTRIBUTE, errorData);	
			}
			enclosedResponse.put(ERROR_ATTRIBUTE, error);
			setJsonrpcVersion("2.0");
		} catch (ConversionException e) {
			e.printStackTrace();
		}
		
	}

	
	private void setResponse(JsTypeComplex jsonRes) throws ConversionException{
		JsTypeObject jsObj = null;
		if (jsonRes != null) { //Validate required parts!
			boolean hasErr=false, hasRes=false;
			for(String key: jsonRes.keys()){
				jsObj = jsonRes.get(key);
				if(key.equals(RESPONSEID_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isSimple()) {
						throw new ConversionException("id must be a simple!");
					}
				}else if(key.equals(RESULT_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					hasRes = true;
					if(hasErr){
						throw new ConversionException("response cannot have error as result at the same time !");
					}
				}else if(key.equals(JSONRPC_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isSimple()) {
						throw new ConversionException("jsonrpc version must be a simple!");
					}
				}else if(key.equals(ERROR_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isComplex()) {
						throw new ConversionException("error must be must be a complex one!");
					}
					JsTypeComplex err = (JsTypeComplex) jsObj;
					if(! err.has(ERRORCODE_ATTRIBUTE) && ! err.has(ERRORMESSAGE_ATTRIBUTE) && ! err.has(ERRORNAME_ATTRIBUTE)){
						throw new ConversionException("error must contain code, message and name attributes at least!");
					}
					hasErr = true;
					if(hasRes){
						throw new ConversionException("response cannot have error as result at the same time !");
					}
				}else if(key.equals(HEADER_ATTRIBUTE)){
					if(jsObj.isNull()) continue;
					if(!jsObj.isComplex()) {
						throw new ConversionException("header must be must be a complex one!");
					}
				}else{
					throw new ConversionException("Message contains illegal attribute name "+key);
				}
					
			}
			this.enclosedResponse = jsonRes;
		} else {
			throw new ConversionException("Empty request in");
		}

	}
	
	public boolean hasResult() {
		return enclosedResponse != null && enclosedResponse.has(RESULT_ATTRIBUTE);
	}

	public boolean hasHeaders() {
		if( enclosedResponse == null || ! enclosedResponse.has(HEADER_ATTRIBUTE)) return false;
		return getHeader().size()>0;
	}

	public boolean hasError() {
		return enclosedResponse != null && enclosedResponse.has(ERROR_ATTRIBUTE);
	}

	public String getResponseId() {
		return enclosedResponse == null || ! enclosedResponse.has(RESPONSEID_ATTRIBUTE)? null : enclosedResponse.get(RESPONSEID_ATTRIBUTE).toStringValue();
	}

	public JsTypeComplex getError() {
		return  (enclosedResponse == null || ! enclosedResponse.has(ERROR_ATTRIBUTE) ? null :(JsTypeComplex) enclosedResponse.get(ERROR_ATTRIBUTE));
	}
	
	public String getErrorName() {
		return getError() == null ? null : getError().get(ERRORNAME_ATTRIBUTE).toStringValue();
	}

	public String getErrorMessage() {
		return getError() == null ? null : getError().get(ERRORMESSAGE_ATTRIBUTE).toStringValue();
	}

	public Integer getErrorCode() {
		return getError() == null ? null : Integer.valueOf(getError().get(ERRORCODE_ATTRIBUTE).toStringValue());
	}

	public JsTypeObject getErrorData() {
		return getError() == null ? null : getError().get(ERRORDATA_ATTRIBUTE);
	}
	
	public JsTypeObject getResult() {
		return enclosedResponse == null || ! enclosedResponse.has(RESULT_ATTRIBUTE) ? null :enclosedResponse.get(RESULT_ATTRIBUTE);
	}

	public <T> T getResultAsJava(Class<? extends T> clazz) throws ConversionException {
		JsonConverter converter = new JsonConverter();
		return converter.toJava(clazz,getResult());
	}
	
	public Object getResultAsJava() throws ConversionException {
		JsonConverter converter = new JsonConverter();
		return converter.toJava(getResult());
	}
	
	public void setJsonrpcVersion(String jsonrpcVersion) {
		try {
			enclosedResponse.putObject(JSONRPC_ATTRIBUTE, new JsTypeSimple(JsTypeObject.quote(jsonrpcVersion),null));
		} catch (ConversionException e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	public String getJsonrpcVersion() {
		if(enclosedResponse.has(JSONRPC_ATTRIBUTE)){
			return   enclosedResponse.get(JSONRPC_ATTRIBUTE).toStringValue();
		}else{
			return null;
		}
	}
	
	void setHeader(Map<String, Object> header) throws
			ConversionException {
		if (header == null){
			enclosedResponse.remove(HEADER_ATTRIBUTE);
			return;
		}
		JsTypeObject jsHdr = converter.toTargetType(header);
		enclosedResponse.put(HEADER_ATTRIBUTE, jsHdr);
	}

	public Map<String, Object> getHeader() {
		if(! enclosedResponse.has(HEADER_ATTRIBUTE) || enclosedResponse.get(HEADER_ATTRIBUTE).isNull()) return new HashMap<String, Object>(0);;
		JsTypeComplex header = (JsTypeComplex) enclosedResponse.get(HEADER_ATTRIBUTE);
		try {
			@SuppressWarnings("unchecked")
			Map<String, Object> jheader = (Map<String, Object>) converter.toJava(HashMap.class,header);
			return jheader;
		} catch (ConversionException e) {
			e.printStackTrace();
			return new HashMap<String, Object>(0);
		}
	}

	public void putToHeader(String fieldName, Object value)
			throws ConversionException  {
		Map<String, Object> header = getHeader();
		if (header == null){
			header = new HashMap<String, Object>();
		}
		header.put(fieldName, value);
		enclosedResponse.put(HEADER_ATTRIBUTE, converter.toTargetType(header));
	}

	public void removeFromHeader(String fieldName) throws 
			ConversionException {
		JsTypeComplex header = (JsTypeComplex) enclosedResponse.get(HEADER_ATTRIBUTE);
		if (header == null) return;
		header.remove(fieldName);
		enclosedResponse.put(HEADER_ATTRIBUTE, header);
	}

	public JsTypeComplex toJSON() {
		return enclosedResponse;
	}

	@Override
	public String toString() {
		return enclosedResponse == null ? null : enclosedResponse.toString();
	}

}
